/* ignuit - Educational software for the GNOME, following the Leitner
 * flash-card system.
 *
 * Copyright (C) 2008 Timothy Musson
 *
 * <trmusson@ihug.co.nz>
 * http://homepages.ihug.co.nz/~trmusson/programs.html#ignuit
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>
#include <gnome.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include "main.h"
#include "prefs.h"
#include "app-window.h"
#include "dialog-properties.h"
#include "dialog-preferences.h"
#include "dialog-editor.h"
#include "dialog-quiz.h"
#include "dialog-about.h"
#include "fileio.h"


typedef struct {

    File             *file;
    Prefs            *prefs;

    Ignuit           *ig;

    GladeXML         *gui;
    GtkWidget        *window;

    GnomeAppBar      *appbar;

    GtkWidget        *scrollw_categories;
    GtkTreeView      *treev_categories;
    GtkTreeStore     *ts_categories;
    GtkTreeSelection *sel_categories;

    GtkWidget        *scrollw_cards;
    GtkTreeView      *treev_cards;
    GtkListStore     *ls_cards;
    GtkTreeSelection *sel_cards;

    GtkWidget        *t_save;
    GtkWidget        *t_quiz_1;

    GtkWidget        *m_save;
    GtkWidget        *m_reset_stats;

    GtkWidget        *m_edit_card;
    GtkWidget        *m_add_card;
    GtkWidget        *m_remove_card;

    GtkWidget        *m_remove_category;

    GtkWidget        *m_quiz_1;
    GtkWidget        *m_quiz_2;
    GtkWidget        *m_quiz_reverse;

    GtkWidget        *r_quiz_card_selection;
    GSList           *quiz_card_group;

    GtkWidget        *b_remove_category;
    GtkWidget        *b_edit_card;
    GtkWidget        *b_add_card;
    GtkWidget        *b_remove_card;

    GtkWidget        *hpaned;

    GtkTreeViewColumn *treev_col_card_front;

    GList            *drag_list;

} WindowData;


/* Combobox ids for import/export filter selection: */
enum {
    FILTER_CSV = 0,
    FILTER_TSV,
    FILTER_NATIVE,
    FILTER_XSLT
};


/* XXX: I have no idea how drag'n'drop works... */

static GtkTargetEntry target_table[] = {
    { "application/x-ignuit-data", GTK_TARGET_SAME_APP, 1 }
};


static gchar*
str_date (gchar *dest, GDate *date)
{
    /* Fill dest with the date in YYYY-MM-DD format, '\0' terminated.
     * The dest string must be already allocated to at least 11 chars.
     */

    if (date == NULL) {
        dest[0] = '-';
        dest[1] = '\0';
        return dest;
    }

    g_snprintf (dest, 11, "%04d-%02d-%02d",
        g_date_get_year (date),
        g_date_get_month (date),
        g_date_get_day (date));

    return dest;
}


static gchar*
truncated_card_text (const gchar *text)
{
    gchar *nl;

    /* Return a copy of the text, truncated at the first '\n' if present. */

    if ((nl = strchr (text, '\n')) == NULL) {
        return g_strdup (text);
    }

    return g_strndup (text, nl - text);
}


static void
update_appbar (WindowData *data)
{
    Category *cat;
    gchar    *s;
    guint    nf, nc;


    nf = file_get_n_cards (data->file);

    if ((cat = file_get_current_category (data->file))) {

        nc = category_get_n_cards (cat);
        s = g_strdup_printf (" Category contains %d of %d cards", nc, nf);

    }
    else {

        s = g_strdup_printf (" File contains %d cards", nf);

    }

    gnome_appbar_set_status (data->appbar, s);

    g_free (s);
}


static void
app_window_update_sensitivity (WindowData *data)
{
    Category *cat;
    gboolean file_has_cards;
    gboolean category_has_cards;


    cat = file_get_current_category (data->file);

    file_has_cards = (file_get_n_cards (data->file) > 0);
    category_has_cards = (cat && category_get_n_cards (cat) > 0);

    gtk_widget_set_sensitive (data->m_reset_stats, file_has_cards);
    gtk_widget_set_sensitive (data->m_quiz_2, file_has_cards);

    gtk_widget_set_sensitive (data->t_quiz_1, category_has_cards);
    gtk_widget_set_sensitive (data->m_quiz_1, category_has_cards);
    gtk_widget_set_sensitive (data->b_add_card, cat != NULL);
    gtk_widget_set_sensitive (data->m_add_card, cat != NULL);
    gtk_widget_set_sensitive (data->b_remove_category, cat != NULL);
    gtk_widget_set_sensitive (data->m_remove_category, cat != NULL);

    update_appbar (data);
}


static void
expired_cell_func (GtkTreeViewColumn *col,
    GtkCellRenderer *renderer,
    GtkTreeModel    *model,
    GtkTreeIter     *iter,
    gpointer        data)
{
    static gboolean expired;


    /* Indicate expired cards with a red asterisk - otherwise blank. */

    gtk_tree_model_get (model, iter, COLUMN_CARD_EXPIRED, &expired, -1);
    if (expired) {

        g_object_set (renderer, "foreground", "#ff0000", NULL);
        g_object_set (renderer, "text", "*", NULL);

    }
    else {

        g_object_set (renderer, "text", "", NULL);

    }
}


static void
group_cell_func (GtkTreeViewColumn *col,
    GtkCellRenderer *renderer,
    GtkTreeModel    *model,
    GtkTreeIter     *iter,
    gpointer        data)
{
    static gchar s[2];
    static Group group;


    gtk_tree_model_get (model, iter, COLUMN_CARD_GROUP, &group, -1);

    s[0] = (group == GROUP_NEW) ? 'N' : '0' + group;
    s[1] = '\0';

    g_object_set (renderer, "text", s, NULL);
}


static void
refresh_category_list (WindowData *data)
{
    GList *cur;
    GtkTreeIter toplevel, child;
    Category *cat;


    /* Clear and re-fill the category treeview. */

    gtk_tree_store_clear (data->ts_categories);

    gtk_tree_store_append (data->ts_categories, &toplevel, NULL);
    gtk_tree_store_set (data->ts_categories, &toplevel,
        COLUMN_CATEGORY_DATA, NULL,
        COLUMN_CATEGORY_TITLE, file_get_title (data->file),
        -1);

    for (cur = file_get_categories(data->file); cur != NULL; cur = cur->next) {

        cat = CATEGORY(cur);

        gtk_tree_store_append (data->ts_categories, &child, &toplevel);
        gtk_tree_store_set (data->ts_categories, &child,
            COLUMN_CATEGORY_DATA, cur,
            COLUMN_CATEGORY_TITLE, category_get_title (cat),
            -1);
    }
}


void
app_window_refresh_card_view (GtkTreeView *treev, GtkListStore *ls,
    GList *cards)
{
    static gchar       crt[11], tst[11], exp[11];
    gchar              *front;
    static GtkTreeIter iter;
    GList              *cur;
    GtkTreeModel       *model;
    Card               *c;


    /* Clear and re-fill the card treeview. */


    /* For speed, temporarily detach model from view. */
    model = gtk_tree_view_get_model(treev);
    g_object_ref(model);
    gtk_tree_view_set_model(treev, NULL);

    gtk_list_store_clear (ls);

    for (cur = cards; cur != NULL; cur = cur->next) {

        c = CARD(cur);

        crt[0] = '\0'; /* str_date (crt, card_get_gdate_created (c)); */
        str_date (tst, card_get_gdate_tested (c));
        str_date (exp, card_get_gdate_expiry (c));

        front = truncated_card_text (card_get_front (c));

        gtk_list_store_append (ls, &iter);
        gtk_list_store_set (ls, &iter,
            COLUMN_CARD_DATA, cur,
            COLUMN_CARD_FRONT, front,
            COLUMN_CARD_BACK, NULL,
            COLUMN_CARD_GROUP, card_get_group (c),
            COLUMN_CARD_CREATED, crt,
            COLUMN_CARD_TESTED, tst,
            COLUMN_CARD_EXPIRED, card_get_expired (c),
            COLUMN_CARD_EXPIRY, exp,
            -1);

        g_free (front);
    }

    /* Re-attach model to view. */
    gtk_tree_view_set_model(treev, model);
    g_object_unref(model);
}


GList*
app_window_find_card_iter (GtkTreeView *treev,
    GtkTreeIter *iter, GList *search)
{
    GtkTreeModel *model;
    gboolean valid;
    GList *item = NULL;


    /* Find the search item in the card treeview, and set iter to it.
     * Return the found item, or NULL.
     */

    model = gtk_tree_view_get_model (treev);
    valid = gtk_tree_model_get_iter_first (model, iter);

    while (valid) {

        gtk_tree_model_get (model, iter, COLUMN_CARD_DATA, &item, -1);
        if (CARD(item) == CARD(search)) { break; }
        valid = gtk_tree_model_iter_next (model, iter);

    }
    return item;
}


void
app_window_set_card_iter (GtkTreeView *treev,
    GtkListStore *ls, GtkTreeIter *iter, GList *item)
{
    gchar        crt[11], tst[11], exp[11];
    gchar        *front;
    GtkTreeModel *model;
    Card         *c;


    /* Link this iter to the given item. */

    c = CARD(item);

    crt[0] = '\0'; /* str_date (crt, card_get_gdate_created (c)); */
    str_date (tst, card_get_gdate_tested (c));
    str_date (exp, card_get_gdate_expiry (c));

    front = truncated_card_text (card_get_front (c));

    model = gtk_tree_view_get_model (treev);
    gtk_list_store_set (ls, iter,
        COLUMN_CARD_DATA, item,
        COLUMN_CARD_FRONT, front,
        COLUMN_CARD_BACK, NULL,
        COLUMN_CARD_GROUP, card_get_group (c),
        COLUMN_CARD_CREATED, crt,
        COLUMN_CARD_TESTED, tst,
        COLUMN_CARD_EXPIRED, card_get_expired (c),
        COLUMN_CARD_EXPIRY, exp,
        -1);

    g_free (front);
}


void
app_window_add_card_iter (GtkTreeView *treev, GtkListStore *ls,
    GtkTreeIter *iter, GList *item)
{
    GtkTreeModel *model;
    GtkTreeIter sibling;


    /* Add new card to the treeview, and select it. */

    model = gtk_tree_view_get_model (treev);
    if (gtk_tree_model_get_iter_first (model, &sibling)) {
        gtk_list_store_insert_before (ls, iter, &sibling);
    }
    else {
        gtk_list_store_append (ls, iter);
    }
    app_window_set_card_iter (treev, ls, iter, item);
    app_window_select_card (treev, item);
}


void
app_window_refresh_card_iter (GtkTreeView *treev,
    GtkListStore *ls, GList *item)
{
    GtkTreeIter iter;

    /* Make sure this row is displaying current information. */

    app_window_find_card_iter (treev, &iter, item);
    app_window_set_card_iter (treev, ls, &iter, item);
}


void
app_window_update_category_root_title (GtkTreeView *treev,
    GtkTreeStore *ts,
    const gchar *title)
{
    GtkTreeIter iter;
    GtkTreeModel *model;

    model = gtk_tree_view_get_model (treev);
    gtk_tree_model_get_iter_first (model, &iter);

    gtk_tree_store_set (ts, &iter,
        COLUMN_CATEGORY_DATA, NULL,
        COLUMN_CATEGORY_TITLE, title,
        -1);
}


void
app_window_select_iter (GtkTreeView *treev, GtkTreeIter *iter)
{
    GtkTreeSelection *sel;
    GtkTreeModel *model;
    GtkTreePath *path;


    /* Select and scroll to an iter in the card TreeView. */

    sel = gtk_tree_view_get_selection (treev);
    gtk_tree_selection_unselect_all (sel);

    model = gtk_tree_view_get_model (treev);
    path = gtk_tree_model_get_path (model, iter);

    gtk_tree_selection_select_iter (sel, iter);
    gtk_tree_view_scroll_to_cell (treev, path, NULL, TRUE, 0.5, 0.0);
}


gboolean
app_window_select_card (GtkTreeView *treev, GList *item)
{
    GtkTreeIter iter;
    gboolean ok;


    /* Select and scroll to an item in the card TreeView. */

    ok = (app_window_find_card_iter (treev, &iter, item) != NULL);
    if (ok) {
        app_window_select_iter (treev, &iter);
    }
    return ok;
}


void
app_window_select_treev_root (GtkTreeView *treev)
{
    GtkTreeSelection *sel;
    GtkTreePath *path;

    path = gtk_tree_path_new_from_string ("0");
    gtk_tree_view_expand_row (treev, path, FALSE);

    sel = gtk_tree_view_get_selection (treev);
    gtk_tree_selection_select_path (sel, path);

    gtk_tree_path_free (path);
}


static void
save_app_window_geometry (WindowData *data)
{
    gint w, h;

    /* Called just before exit to store window size, etc. */

    gtk_window_get_size (GTK_WINDOW(data->window), &w, &h);
    prefs_set_app_size (data->prefs, w, h);

    w = gtk_paned_get_position (GTK_PANED(data->hpaned));
    prefs_set_category_pane_width (data->prefs, w);

    w = gtk_tree_view_column_get_width (data->treev_col_card_front);
    prefs_set_card_column_front_width (data->prefs, w);
}


static gboolean
get_save_as_filename (WindowData *data)
{
    GtkWidget *dialog;
    gchar *fname_new;
    const gchar *fname_old;
    gint result;


    dialog = gtk_file_chooser_dialog_new (
        _("Save As"), GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_SAVE,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    fname_old = file_get_filename (data->file);

    if (fname_old != NULL && g_path_is_absolute (fname_old)) {
        gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(dialog), fname_old);
    }
    else if (prefs_get_workdir (data->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->prefs));
    }
    result = gtk_dialog_run (GTK_DIALOG(dialog));

    fname_new = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname_new != NULL) {
        file_set_filename (data->file, fname_new);
        g_free (fname_new);
        return TRUE;
    }
    return FALSE;
}


static void
error_dialog (GtkWindow *parent, const gchar *message, GError *err)
{
    GtkDialog *dialog;
    gchar *msg;


    if (err && err->message) {
        msg = g_strdup_printf ("%s\n\n%s", message, err->message);
    }
    else {
        msg = g_strdup (message);
    }

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_CLOSE,
        msg);

    gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    g_free (msg);
}


static gint
ask_save (GtkWindow *parent)
{
    GtkDialog *dialog;
    gint result;

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_QUESTION,
        GTK_BUTTONS_NONE,
        _("<b>The current file has not been saved</b>\n\nChanges since last save will be lost."));

    gtk_dialog_add_buttons (dialog,
        _("Close _without Saving"), GTK_RESPONSE_REJECT,
        GTK_STOCK_CANCEL,           GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE,             GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (dialog, GTK_RESPONSE_ACCEPT);
    result = gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    if (result != GTK_RESPONSE_ACCEPT && result != GTK_RESPONSE_REJECT) {
        result = GTK_RESPONSE_CANCEL;
    }

    return result;
}


gboolean
ask_yes_or_no (GtkWindow *parent, const gchar *question,
    GtkResponseType def)
{
    GtkDialog *dialog;
    gint result;

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_QUESTION,
        GTK_BUTTONS_YES_NO,
        question);

    gtk_dialog_set_default_response (dialog, def);
    result = gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    return result == GTK_RESPONSE_YES;
}


static void
unable_to_save_file (GtkWindow *parent, GError *err)
{
    error_dialog (parent, _("<b>Unable to save file</b>"), err);

     if (err != NULL) {
        g_error_free (err);
    }
}


static gboolean
save_changes (WindowData *data)
{
    GError *err = NULL;
    gint reply;


    /* If the file has been changed, ask whether or not to save it.
     * Return TRUE only if the changes are either unwanted or
     * successfully saved.
     */

    if (!file_get_changed (data->file)) {
        return TRUE;
    }

    reply = ask_save (GTK_WINDOW(data->window));

    if (reply == GTK_RESPONSE_CANCEL) {
        return FALSE;
    }

    if (reply == GTK_RESPONSE_REJECT) {
        return TRUE;
    }

    if (!file_get_filename (data->file) && !get_save_as_filename (data)) {
        return FALSE;
    }

    if (!fileio_save (data->file, file_get_filename (data->file), &err)) {

        unable_to_save_file (GTK_WINDOW(data->window), err);

        return FALSE;
    }

    return TRUE;
}


static void
cb_m_new (GtkWidget *widget, WindowData *data)
{
    /* Close the current file and create a new one. */

    dialog_editor_check_changed ();

    if (!save_changes (data)) { return; }

    dialog_editor_kill ();

    file_free (data->ig->file, TRUE);

    data->ig->file = file_new ();
    data->file = data->ig->file;
    file_set_current_category (data->file, NULL);

    refresh_category_list (data);
    app_window_refresh_card_view (data->treev_cards, data->ls_cards,
        file_get_current_category_cards (data->file));

    app_window_select_treev_root (data->treev_categories);
}


static void
cb_m_open (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    File *file;
    gint result;
    gchar *fname;

    /* Prompt for a file to open, and try to load it. If successful, close
     * the current file and display the newly loaded one.
     */

    dialog_editor_check_changed ();

    if (!save_changes (data)) { return; }

    dialog_editor_kill ();

    dialog = gtk_file_chooser_dialog_new (_("Open"),
        GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
        NULL);
    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->prefs));
    }

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        if ((file = fileio_load (fname, &err)) != NULL) {

            /* Success - replace the current file with the opened one. */

            file_set_current_category (file, NULL);
            file_free (data->ig->file, TRUE);
            data->ig->file = file;
            data->file = file;

            if (g_path_is_absolute (fname)) {
                gchar *workdir;
                workdir = g_path_get_dirname (fname);
                prefs_set_workdir (data->prefs, workdir);
                g_free (workdir);
            }

            refresh_category_list (data);
            app_window_select_treev_root (data->treev_categories);
            app_window_refresh_card_view (data->treev_cards, data->ls_cards,
                file_get_current_category_cards (data->file));

        }
        else {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to load file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }
        g_free (fname);
    }
}


static void
cb_m_save (GtkWidget *widget, WindowData *data)
{
    GError *err = NULL;

    if (file_get_filename (data->file) == NULL) {
        if (!get_save_as_filename (data)) {
            return;
        }
    }
    if (!fileio_save (data->file, file_get_filename (data->file), &err)) {

        unable_to_save_file (GTK_WINDOW(data->window), err);

    }
}


static void
cb_m_save_as (GtkWidget *widget, WindowData *data)
{
    if (get_save_as_filename (data)) {

        GError *err = NULL;

        if (!fileio_save (data->file, file_get_filename (data->file), &err)) {

            unable_to_save_file (GTK_WINDOW(data->window), err);

        }
    }
}


static GList*
populate_import_export_combo_box (WindowData *data, GtkComboBox *combo,
    gboolean is_import)
{
    GList       *list = NULL;
    GDir        *dir;
    const gchar *whichdir;
    gchar       *dname;

    whichdir = is_import ? IMPORT_DIR : EXPORT_DIR;

    dname = gnome_program_locate_file (data->ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, whichdir, TRUE, NULL);

    if (!dname) {
        g_warning ("Can't find directory: %s\n", whichdir);
        return NULL;
    }

    gtk_combo_box_append_text (combo, _("CSV"));
    gtk_combo_box_append_text (combo, _("TSV"));
    gtk_combo_box_append_text (combo, "Ignuit");

    dir = g_dir_open (dname, 0, NULL);
    if (dir) {

        const gchar *entry;

        while ((entry = g_dir_read_name (dir))) {

            gchar *s, *ext;

            s = g_filename_to_utf8 (entry, -1, NULL, NULL, NULL);
            if (s != NULL) {
                if ((ext = strchr (s, '.')) != NULL) {
                    *ext = '\0';
                }
                gtk_combo_box_append_text (combo, s);
                g_free (s);
                list = g_list_append (list, g_strdup (entry));
            }
            else {
                g_warning ("g_filename_to_utf8 failed for '%s'\n", entry);
            }
        }

        g_dir_close (dir);
    }
    g_free (dname);

    return list;
}


static void
cb_m_import (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    GtkWidget *box;
    GtkWidget *combo;
    GtkWidget *label;
    GList     *filters;
    File      *file;
    gint      result, combo_selection;
    gchar     *fname;


    /* Import a non-native file, adding it as a new category to the
     * currently open file.
     */

    dialog_editor_check_changed ();
    dialog_editor_kill ();

    dialog = gtk_file_chooser_dialog_new (_("Import"),
        GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
        NULL);
    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->prefs));
    }

    box = gtk_hbox_new (FALSE, 0);

    combo = gtk_combo_box_new_text ();
    filters = populate_import_export_combo_box (data,
        GTK_COMBO_BOX(combo), TRUE);
    gtk_combo_box_set_active (GTK_COMBO_BOX(combo), 0);

    gtk_box_pack_end (GTK_BOX(box), combo, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic (_("Import _Filter:"));
    gtk_box_pack_end (GTK_BOX(box), label, FALSE, FALSE, 12);

    gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER(dialog), box);
    gtk_widget_show_all (box);

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    combo_selection = gtk_combo_box_get_active (GTK_COMBO_BOX(combo));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        const gchar *fname_filter;

        switch (combo_selection) {
        case FILTER_CSV:
            file = fileio_import_sv (fname, ",", &err);
            break;
        case FILTER_TSV:
            file = fileio_import_sv (fname, "\t", &err);
            break;
        case FILTER_NATIVE:
            file = fileio_load (fname, &err);
            break;
        default:
            /* >= FILTER_XSLT */
            fname_filter = g_list_nth_data (filters,
                combo_selection - FILTER_XSLT);
            file = fileio_import (data->ig, fname, fname_filter, &err);
            break;
        }

        if (file) {

            Category *cat, *cat_tmp;
            GtkTreeIter root, child;
            GList *cur, *item;
     
            if (g_path_is_absolute (fname)) {
                gchar *workdir;
                workdir = g_path_get_dirname (fname);
                prefs_set_workdir (data->prefs, workdir);
                g_free (workdir);
            }

            /* Add cards from the imported file to the currently open file. */

            /* XXX: This copies cards into one big category, ignoring
             * imported categories.
             */

            cat_tmp = file_get_current_category (data->file);

            cat = category_new (_("Imported Cards"));
            item = file_add_category (data->file, cat);

            file_set_current_category (data->file, cat);
            file_set_changed (data->file, TRUE);

            for (cur = file_get_cards (file); cur != NULL; cur = cur->next) {
                file_add_loaded_card (data->file, CARD(cur));
            }

            /* Update the category TreeView. */

            gtk_tree_model_get_iter_first (GTK_TREE_MODEL(data->ts_categories),
                &root);
            gtk_tree_store_append (data->ts_categories, &child, &root);
            gtk_tree_store_set (data->ts_categories, &child,
                COLUMN_CATEGORY_DATA, item,
                COLUMN_CATEGORY_TITLE, category_get_title (cat),
                -1);

            gtk_tree_view_expand_all (data->treev_categories);

            file_set_current_category (data->file, cat_tmp);

            /* Free the imported file, but not the cards. */
            file_free (file, FALSE);

            update_appbar (data);
        }
        else {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to import file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }

    }

    if (filters) {
        g_list_foreach (filters, (GFunc)g_free, NULL);
        g_list_free (filters);
    }
}


static void
cb_m_export (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    GtkWidget *box;
    GtkWidget *combo;
    GtkWidget *label;
    GList     *filters;
    gchar     *fname;
    gint      result, combo_selection;


    dialog = gtk_file_chooser_dialog_new (
        _("Export"), GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_SAVE,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->prefs));
    }

    box = gtk_hbox_new (FALSE, 0);

    combo = gtk_combo_box_new_text ();
    filters = populate_import_export_combo_box (data,
        GTK_COMBO_BOX(combo), FALSE);
    gtk_combo_box_set_active (GTK_COMBO_BOX(combo), 0);

    gtk_box_pack_end (GTK_BOX(box), combo, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic (_("Export _Filter:"));
    gtk_box_pack_end (GTK_BOX(box), label, FALSE, FALSE, 12);

    gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER(dialog), box);
    gtk_widget_show_all (box);

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    combo_selection = gtk_combo_box_get_active (GTK_COMBO_BOX(combo));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        const gchar *filter;
        gboolean ok, changed_tmp;

        switch (combo_selection) {
        case FILTER_CSV:
            ok = fileio_export_sv (data->file, fname, ",", &err);
            break;
        case FILTER_TSV:
            ok = fileio_export_sv (data->file, fname, "\t", &err);
            break;
        case FILTER_NATIVE:
            changed_tmp = file_get_changed (data->file);
            ok = fileio_save (data->file, fname, &err);
            file_set_changed (data->file, changed_tmp);
            break;
        default:
            /* >= FILTER_XSLT */
            filter = g_list_nth_data (filters,
                combo_selection - FILTER_XSLT);
            ok = fileio_export (data->ig, fname, filter, &err);
            break;
        }

        if (!ok) {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to export file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }

        g_free (fname);

    }

    if (filters) {
        g_list_foreach (filters, (GFunc)g_free, NULL);
        g_list_free (filters);
    }
}


static void
cb_m_reset_stats (GtkWidget *widget, WindowData *data)
{
    /* Forget the dates and counts for all cards in the file. */

    if (!ask_yes_or_no (GTK_WINDOW(data->window),
        _("<b>Expiry dates will be reset</b>\n\nRecorded dates and statistics for all cards in the current file will be lost. Continue anyway?"),
        GTK_RESPONSE_NO)) {
        return;
    }

    file_reset_statistics (data->file);
    file_set_changed (data->file, TRUE);

    app_window_refresh_card_view (data->treev_cards, data->ls_cards,
        file_get_current_category_cards (data->file));
}


static void
cb_m_properties (GtkWidget *widget, WindowData *data)
{
    dialog_properties (data->ig, data->treev_categories, data->ts_categories);
}


static void
cb_m_quit (GtkWidget *widget, WindowData *data)
{
    dialog_editor_check_changed ();

    if (!save_changes (data)) {return;}

    save_app_window_geometry (data);

    g_object_unref (G_OBJECT(data->gui));
    g_free (data);

    gtk_main_quit ();
}


static void
cb_m_add_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeIter iter;
    GList *item;
    Card *c;


    /* Make a new blank card, add it to the currently selected category,
     * and open the card editor to fill it in.
     */

    dialog_editor_check_changed ();

    c = card_new ();
    card_set_front (c, "");
    card_set_back (c, "");
    file_add_card (data->file, file_get_current_category (data->file), c);
    file_set_changed (data->file, TRUE);

    item = file_get_current_category_cards (data->file);
    file_set_current_item (data->file, item);

    app_window_add_card_iter (data->treev_cards, data->ls_cards,
        &iter, item);

    app_window_update_sensitivity (data);

    dialog_editor (data->ig, data->window,
        data->treev_cards, data->ls_cards);
}


static void
cb_m_remove_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;
    Card *c;


    /* Remove currently selected cards from the file. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    if (!ask_yes_or_no (GTK_WINDOW(data->window),
        _("<b>Permanently remove the selected cards?</b>"),
        GTK_RESPONSE_NO)) {
        g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
        g_list_free (rows);
        return;
    }

    /* XXX: nicer to do it without closing the editor window */
    dialog_editor_check_changed ();
    dialog_editor_kill ();

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (GTK_TREE_MODEL(data->ls_cards),
                &iter, path)) {

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
                gtk_list_store_remove(data->ls_cards, &iter);

                c = CARD(item);
                file_remove_card (data->file, c);
                file_set_current_item (data->file, NULL);
                card_free (c);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    file_set_changed (data->file, TRUE);

    app_window_update_sensitivity (data);

    /* What fun! */
}


static void
cb_m_edit_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GList *rows, *item;


    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    g_assert (rows != NULL);

    path = (GtkTreePath*)rows->data; /* Use the first selected row. */

    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
    app_window_select_iter (data->treev_cards, &iter);

    file_set_current_item (data->file, item);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    dialog_editor (data->ig, data->window,
        data->treev_cards, data->ls_cards);
}


static void
cb_m_add_category (GtkWidget *widget, WindowData *data)
{
    GtkTreeIter root, child;
    Category *cat;
    GList *item;


    /* Create a new category and add it to the file. */

    cat = category_new (NULL);
    item = file_add_category (data->file, cat);
    file_set_changed (data->file, TRUE);

    /* Update the category TreeView. */

    gtk_tree_model_get_iter_first (GTK_TREE_MODEL(data->ts_categories), &root);
    gtk_tree_store_append (data->ts_categories, &child, &root);
    gtk_tree_store_set (data->ts_categories, &child,
        COLUMN_CATEGORY_DATA, item,
        COLUMN_CATEGORY_TITLE, category_get_title (cat),
        -1);

    /* Make sure the new category is visible. */

    gtk_tree_view_expand_all (data->treev_categories);


    /* XXX: select the new category */
}


static void
cb_m_remove_category (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GList *item;
    Category *cat;


    /* Delete the selected category (and any cards it contains). */

    if (gtk_tree_selection_get_selected (data->sel_categories, &model,
        &iter)) {

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
        if (item == NULL) { return; }

        cat = CATEGORY(item);
        if (category_get_n_cards (cat) && !ask_yes_or_no
            (GTK_WINDOW(data->window),
            _("<b>Delete this non-empty category?</b>\n\nThe category and its contents will be permanently deleted."),
            GTK_RESPONSE_NO)) {
            return;
        }

        dialog_editor_check_changed ();
        dialog_editor_kill ();

        file_remove_category (data->file, cat);
        file_set_current_category (data->file, NULL);
        file_set_current_item (data->file, NULL);
        file_set_changed (data->file, TRUE);

        gtk_tree_store_remove (data->ts_categories, &iter);

        app_window_refresh_card_view (data->treev_cards, data->ls_cards,
            file_get_current_category_cards (data->file));

        app_window_update_sensitivity (data);
    }
}


static void
cb_m_preferences (GtkWidget *widget, WindowData *data)
{
    dialog_preferences (data->ig);
}


static gint
get_quiz_card_selection (WindowData *data)
{
    GSList *cur;
    gint i = QUIZ_N_CARD_SELECTIONS;

    for (cur = data->quiz_card_group; cur != NULL; cur = cur->next) {

        i--;

        if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM(cur->data))) {
            return i;
        }
    }
    g_assert_not_reached ();
}


static void
cb_m_quiz_1 (GtkWidget *widget, WindowData *data)
{
    QuizInfo info;

    /* Quiz the currently selected category. */

    dialog_editor_check_changed ();
    dialog_editor_kill ();

    info.category_selection = QUIZ_CURRENT_CATEGORY;
    info.card_selection = get_quiz_card_selection (data);
    info.reverse = gtk_check_menu_item_get_active
        (GTK_CHECK_MENU_ITEM(data->m_quiz_reverse));

    dialog_quiz (data->ig, data->treev_cards, data->ls_cards, &info);
}


static void
cb_m_quiz_2 (GtkWidget *widget, WindowData *data)
{
    QuizInfo info;

    /* Quiz all cards in the file. */

    dialog_editor_check_changed ();
    dialog_editor_kill ();

    info.category_selection = QUIZ_ALL_CATEGORIES;
    info.card_selection = get_quiz_card_selection (data);
    info.reverse = gtk_check_menu_item_get_active
        (GTK_CHECK_MENU_ITEM(data->m_quiz_reverse));

    dialog_quiz (data->ig, data->treev_cards, data->ls_cards, &info);
}


static void
cb_m_about (GtkWidget *widget, WindowData *data)
{
    dialog_about (data->window);
}


static void
tweak_editor (WindowData *data)
{
    if (file_get_current_category_cards (data->file)) {
        app_window_select_treev_root (data->treev_cards);
    }
    dialog_editor_tweak ();
}


static void
cb_treev_category_selection (GtkTreeSelection *sel, WindowData *data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;


    /* A category has been selected in the category TreeView.
     * Update the card TreeView accordingly.
     */

    if (gtk_tree_selection_get_selected (sel, &model, &iter)) {

        GtkAdjustment *adj;
        GList *item;
        Category *cat;

        dialog_editor_check_changed ();

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
        cat = item ? CATEGORY(item) : NULL;

        file_set_current_category (data->file, cat);
        file_set_current_item (data->file,
            cat ? category_get_cards (cat) : NULL);

        adj = gtk_scrolled_window_get_vadjustment
            (GTK_SCROLLED_WINDOW(data->scrollw_cards));
        gtk_adjustment_set_value (adj, 0);

        app_window_refresh_card_view (data->treev_cards, data->ls_cards,
            file_get_current_category_cards (data->file));

        tweak_editor (data);

        app_window_update_sensitivity (data);
    }
}


static void
cb_treev_category_title_edited (GtkCellRendererText *cell,
    gchar *path_string, gchar *new_text, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;


    /* A category's title has been edited in the category TreeView. */

    model = gtk_tree_view_get_model (data->treev_categories);
    path = gtk_tree_path_new_from_string (path_string);

    if (gtk_tree_model_get_iter (model, &iter, path)) {

        GList *item;

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);

        if (item != NULL) {
            /* It's an ordinary category. */
            category_set_title (CATEGORY(item), new_text);
        }
        else {
            /* It's the root category. */
            file_set_title (data->file, new_text);
        }

        gtk_tree_store_set (data->ts_categories, &iter,
            COLUMN_CATEGORY_TITLE, new_text, -1);

    }
    gtk_tree_path_free (path);
}


static void
cb_treev_card_activated (GtkTreeView *treev, GtkTreePath *path,
    GtkTreeViewColumn *column, WindowData *data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;

    /* A row in the card TreeView has been double-clicked. */

    model = gtk_tree_view_get_model (treev);
    if (gtk_tree_model_get_iter (model, &iter, path)) {

        /* Make this the only selected card. */

        GList *item;

        dialog_editor_check_changed ();

        gtk_tree_selection_unselect_all (data->sel_cards);
        gtk_tree_selection_select_iter (data->sel_cards, &iter);
        gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
        file_set_current_item (data->file, item);

        /* Edit the card. */

        dialog_editor (data->ig, data->window,
            data->treev_cards, data->ls_cards);
    }
}


static gboolean
drag_selected_foreach_func (GtkTreeModel  *model,
    GtkTreePath *path, GtkTreeIter *iter, WindowData *data)
{
    GList *item;

    gtk_tree_model_get (model, iter, COLUMN_CARD_DATA, &item, -1);
    data->drag_list = g_list_prepend (data->drag_list, CARD(item));
    return TRUE;
}


static void
cb_treev_cards_drag_data_get (GtkTreeView *treev,
    GdkDragContext *context, GtkSelectionData *sln_data, guint info,
    guint t, WindowData *data)
{
    GtkTreeModel *model;
    GList *rows;


    /* XXX: this works, but I don't think it's the correct way
     * to do drag and drop */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);

    if (rows != NULL) {

        gtk_tree_selection_selected_foreach (data->sel_cards,
            (GtkTreeSelectionForeachFunc)drag_selected_foreach_func, data);

        g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
        g_list_free (rows);
    }
}


static void
cb_treev_categories_drag_data_received (GtkTreeView *treev,
    GdkDragContext *context, gint x, gint y, GtkSelectionData *sln_data,
    guint info, guint time, WindowData *data)
{
    GList *cur, *item;
    Category *cat_tmp, *cat_dest;
    GtkTreePath *path;
    GtkTreeModel *model;
    GtkTreeIter iter;


    /* XXX: this works, but I don't think it's the correct way
     * to do drag and drop */

    gtk_tree_view_get_dest_row_at_pos (treev, x, y, &path, NULL);
    if (path == NULL || gtk_tree_path_get_depth (path) < 2) {
        g_list_free (data->drag_list);
        data->drag_list = NULL;
        gtk_drag_finish (context, FALSE, FALSE, time);
        return;
    }

    model = gtk_tree_view_get_model (treev);
    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
    cat_dest = CATEGORY(item);

    /* XXX: nicer to do it without closing the editor window */
    dialog_editor_check_changed ();
    dialog_editor_kill ();

    for (cur = data->drag_list; cur != NULL; cur = cur->next) {

        cat_tmp = card_get_category (CARD(cur));
        category_remove_card (cat_tmp, CARD(cur));
        category_prepend_card (cat_dest, CARD(cur));

    }
    g_list_free (data->drag_list);
    data->drag_list = NULL;
    gtk_drag_finish (context, FALSE, FALSE, time);

    file_set_changed (data->file, TRUE);

    /* XXX: it'd be nicer to remove individual items from the card view,
     * rather than refreshing the entire card view.
     */
    app_window_refresh_card_view (data->treev_cards, data->ls_cards,
        file_get_current_category_cards (data->file));
}


static void
cb_treev_cards_selection (GtkTreeSelection *sel, WindowData *data)
{
    GtkTreeModel *model;
    GList *rows;
    gboolean selected;


    rows = gtk_tree_selection_get_selected_rows (sel, &model);
    selected = (rows != NULL);

    gtk_widget_set_sensitive (data->m_edit_card, selected);
    gtk_widget_set_sensitive (data->m_remove_card, selected);

    gtk_widget_set_sensitive (data->b_edit_card, selected);
    gtk_widget_set_sensitive (data->b_remove_card, selected);

    if (rows) {

        g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
        g_list_free (rows);
    }
}


static void
cb_window_destroy (GtkWidget *widget, WindowData *data)
{
    g_object_unref (G_OBJECT(data->gui));
    g_free (data);

    gtk_main_quit ();
}


static gboolean
cb_window_delete (GtkWidget *widget, GdkEvent *event, WindowData *data)
{
    dialog_editor_check_changed ();

    if (!save_changes (data)) {return TRUE;}

    save_app_window_geometry (data);

    return FALSE;
}


static void
add_card_column (WindowData *data, const gchar *title, gint col,
    GtkTreeCellDataFunc func)
{
    GtkCellRenderer *r;
    GtkTreeViewColumn *c;


    /* Add a column to the card TreeView. */

    r = gtk_cell_renderer_text_new ();
    c = gtk_tree_view_column_new_with_attributes (title, r, "text", col, NULL);
    gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN(c), TRUE);
    gtk_tree_view_append_column (data->treev_cards, c);
    if (func != NULL) {
        gtk_tree_view_column_set_cell_data_func (c, r, func, NULL, NULL);
    }

    if (col == COLUMN_CARD_FRONT) {
        gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN(c),
            GTK_TREE_VIEW_COLUMN_FIXED);
        g_object_set (r, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
        data->treev_col_card_front = c;
    }
}


void
app_window (Ignuit *ig)
{
    WindowData *data;
    GtkWidget *m_new, *m_open, *m_save_as, *m_quit;
    GtkWidget *m_import, *m_export;
    GtkWidget *m_add_category, *b_add_category;
    GtkWidget *m_properties, *m_preferences;
    GtkWidget *m_about;
    GtkWidget *t_new, *t_open, *t_properties;
    GtkWidget *w;
    GtkCellRenderer *rend;
    GtkTreeViewColumn *col;
    gchar *gladefile, *iconfile;


    gladefile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_GLADE_MAIN, TRUE, NULL);

    if (gladefile == NULL) {
        g_warning ("Can't find file: %s\n", F_GLADE_MAIN);
        exit (EXIT_FAILURE);
    }

    data = g_new0 (WindowData, 1);
    data->gui = glade_xml_new (gladefile, NULL, NULL);
    g_free (gladefile);

    data->file = ig->file;
    data->prefs = ig->prefs;
    data->ig = ig;

    data->drag_list = NULL;

    ig->app = glade_xml_get_widget (data->gui, "app");

    data->window = ig->app;

    gtk_window_set_default_size (GTK_WINDOW(data->window),
        prefs_get_app_width (data->prefs),
        prefs_get_app_height (data->prefs));

    m_new = glade_xml_get_widget (data->gui, "m_new");
    m_open = glade_xml_get_widget (data->gui, "m_open");
    data->m_save = glade_xml_get_widget (data->gui, "m_save");
    m_save_as = glade_xml_get_widget (data->gui, "m_save_as");
    m_import = glade_xml_get_widget (data->gui, "m_import");
    m_export = glade_xml_get_widget (data->gui, "m_export");
    data->m_reset_stats = glade_xml_get_widget (data->gui, "m_reset_stats");
    m_properties = glade_xml_get_widget (data->gui, "m_properties");
    m_quit = glade_xml_get_widget (data->gui, "m_quit");

    data->m_add_card = glade_xml_get_widget (data->gui, "m_add_card");
    data->m_remove_card = glade_xml_get_widget (data->gui, "m_remove_card");
    data->m_edit_card = glade_xml_get_widget (data->gui, "m_edit_card");
    m_add_category = glade_xml_get_widget (data->gui, "m_add_category");
    data->m_remove_category = glade_xml_get_widget (data->gui, "m_remove_category");
    m_preferences = glade_xml_get_widget (data->gui, "m_preferences");

    data->m_quiz_1 = glade_xml_get_widget (data->gui, "m_quiz_1");
    data->m_quiz_2 = glade_xml_get_widget (data->gui, "m_quiz_2");
    data->m_quiz_reverse = glade_xml_get_widget (data->gui, "m_quiz_reverse");

    data->r_quiz_card_selection = glade_xml_get_widget
        (data->gui, "r_all_cards");
    w = glade_xml_get_widget (data->gui, "r_new_and_expired_cards");

    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM(w), TRUE);

    data->quiz_card_group = gtk_radio_menu_item_get_group
        (GTK_RADIO_MENU_ITEM(data->r_quiz_card_selection));

    m_about = glade_xml_get_widget (data->gui, "m_about");

    t_new = glade_xml_get_widget (data->gui, "t_new");
    t_open = glade_xml_get_widget (data->gui, "t_open");
    data->t_save = glade_xml_get_widget (data->gui, "t_save");
    t_properties = glade_xml_get_widget (data->gui, "t_properties");
    data->t_quiz_1 = glade_xml_get_widget (data->gui, "t_quiz");

    b_add_category = glade_xml_get_widget (data->gui, "b_add_category");
    data->b_remove_category = glade_xml_get_widget (data->gui, "b_remove_category");

    data->b_edit_card = glade_xml_get_widget (data->gui, "b_edit_card");
    data->b_add_card = glade_xml_get_widget (data->gui, "b_add_card");
    data->b_remove_card = glade_xml_get_widget (data->gui, "b_remove_card");

    data->hpaned = glade_xml_get_widget (data->gui, "hpaned");

    data->appbar = (GnomeAppBar*)glade_xml_get_widget (data->gui, "appbar");


    /* category list */

    data->scrollw_categories = glade_xml_get_widget (data->gui,
        "scrollw_categories");
    data->treev_categories = GTK_TREE_VIEW(glade_xml_get_widget (data->gui,
        "treeview_categories"));
    data->ts_categories = gtk_tree_store_new (CATEGORIES_N_COLUMNS,
        G_TYPE_POINTER, G_TYPE_STRING);
    gtk_tree_view_set_model (data->treev_categories,
        GTK_TREE_MODEL(data->ts_categories));
    g_object_unref (data->ts_categories);

    rend = gtk_cell_renderer_text_new ();
    col = gtk_tree_view_column_new_with_attributes (NULL,
        rend, "text", COLUMN_CATEGORY_TITLE, NULL);
    g_object_set (rend, "editable", TRUE, NULL);
    gtk_tree_view_append_column (data->treev_categories, col);

    data->sel_categories = gtk_tree_view_get_selection (data->treev_categories);
    gtk_tree_selection_set_mode (data->sel_categories, GTK_SELECTION_BROWSE);

    g_signal_connect (G_OBJECT(data->sel_categories), "changed",
        G_CALLBACK (cb_treev_category_selection), data);
    g_signal_connect(rend, "edited",
        (GCallback)cb_treev_category_title_edited, data);


    /* card list */

    data->scrollw_cards = glade_xml_get_widget (data->gui, "scrollw_cards");
    data->treev_cards = GTK_TREE_VIEW(glade_xml_get_widget (data->gui,
        "treeview_cards"));

    data->ls_cards = gtk_list_store_new (CARDS_N_COLUMNS,
        G_TYPE_POINTER, /* this card's category GList node */
        G_TYPE_STRING,  /* front */
        G_TYPE_STRING,  /* back */
        G_TYPE_INT,     /* group */
        G_TYPE_STRING,  /* created */
        G_TYPE_STRING,  /* date last tested */
        G_TYPE_BOOLEAN, /* expired */
        G_TYPE_STRING); /* date of expiry */

    gtk_tree_view_set_model (data->treev_cards,
        GTK_TREE_MODEL(data->ls_cards));
    g_object_unref (data->ls_cards);

    add_card_column (data, _("Front"), COLUMN_CARD_FRONT, NULL);
#if 0
    add_card_column (data, _("Back"), COLUMN_CARD_BACK, NULL);
#endif
    add_card_column (data, _("G") /* Group */ , COLUMN_CARD_BACK,
        group_cell_func);
#if 0
    add_card_column (data, _("Added"), COLUMN_CARD_CREATED, NULL);
#endif
    add_card_column (data, _("Tested"), COLUMN_CARD_TESTED, NULL);
    add_card_column (data, _("!") /* Expired! */ , COLUMN_CARD_EXPIRED,
        expired_cell_func);
    add_card_column (data, _("Expiry"), COLUMN_CARD_EXPIRY, NULL);

    data->sel_cards = gtk_tree_view_get_selection (data->treev_cards);
    gtk_tree_selection_set_mode (data->sel_cards, GTK_SELECTION_MULTIPLE);

    g_signal_connect (G_OBJECT(data->sel_cards), "changed",
        G_CALLBACK (cb_treev_cards_selection), data);


    /* signals */

    g_signal_connect (G_OBJECT(data->window), "destroy",
        G_CALLBACK(cb_window_destroy), data);
    g_signal_connect (G_OBJECT(data->window), "delete_event",
        G_CALLBACK(cb_window_delete), data);

    g_signal_connect (G_OBJECT(m_new), "activate",
        G_CALLBACK(cb_m_new), data);
    g_signal_connect (G_OBJECT(m_open), "activate",
        G_CALLBACK(cb_m_open), data);
    g_signal_connect (G_OBJECT(data->m_save), "activate",
        G_CALLBACK(cb_m_save), data);
    g_signal_connect (G_OBJECT(m_save_as), "activate",
        G_CALLBACK(cb_m_save_as), data);
    g_signal_connect (G_OBJECT(m_import), "activate",
        G_CALLBACK(cb_m_import), data);
    g_signal_connect (G_OBJECT(m_export), "activate",
        G_CALLBACK(cb_m_export), data);
    g_signal_connect (G_OBJECT(data->m_reset_stats), "activate",
        G_CALLBACK(cb_m_reset_stats), data);
    g_signal_connect (G_OBJECT(m_properties), "activate",
        G_CALLBACK(cb_m_properties), data);
    g_signal_connect (G_OBJECT(m_quit), "activate",
        G_CALLBACK(cb_m_quit), data);

    g_signal_connect (G_OBJECT(data->m_edit_card), "activate",
        G_CALLBACK(cb_m_edit_card), data);
    g_signal_connect (G_OBJECT(data->m_add_card), "activate",
        G_CALLBACK(cb_m_add_card), data);
    g_signal_connect (G_OBJECT(data->m_remove_card), "activate",
        G_CALLBACK(cb_m_remove_card), data);
    g_signal_connect (G_OBJECT(m_add_category), "activate",
        G_CALLBACK(cb_m_add_category), data);
    g_signal_connect (G_OBJECT(data->m_remove_category), "activate",
        G_CALLBACK(cb_m_remove_category), data);
    g_signal_connect (G_OBJECT(m_preferences), "activate",
        G_CALLBACK(cb_m_preferences), data);

    g_signal_connect (G_OBJECT(data->m_quiz_1), "activate",
        G_CALLBACK(cb_m_quiz_1), data);
    g_signal_connect (G_OBJECT(data->m_quiz_2), "activate",
        G_CALLBACK(cb_m_quiz_2), data);

    g_signal_connect (G_OBJECT(m_about), "activate",
        G_CALLBACK(cb_m_about), data);

    g_signal_connect (G_OBJECT(t_new), "clicked",
        G_CALLBACK(cb_m_new), data);
    g_signal_connect (G_OBJECT(t_open), "clicked",
        G_CALLBACK(cb_m_open), data);
    g_signal_connect (G_OBJECT(data->t_save), "clicked",
        G_CALLBACK(cb_m_save), data);
    g_signal_connect (G_OBJECT(t_properties), "clicked",
        G_CALLBACK(cb_m_properties), data);
    g_signal_connect (G_OBJECT(data->t_quiz_1), "clicked",
        G_CALLBACK(cb_m_quiz_1), data);

    g_signal_connect (G_OBJECT(b_add_category), "clicked",
        G_CALLBACK(cb_m_add_category), data);
    g_signal_connect (G_OBJECT(data->b_remove_category), "clicked",
        G_CALLBACK(cb_m_remove_category), data);

    g_signal_connect (G_OBJECT(data->b_edit_card), "clicked",
        G_CALLBACK(cb_m_edit_card), data);
    g_signal_connect (G_OBJECT(data->b_add_card), "clicked",
        G_CALLBACK(cb_m_add_card), data);
    g_signal_connect (G_OBJECT(data->b_remove_card), "clicked",
        G_CALLBACK(cb_m_remove_card), data);

    g_signal_connect (G_OBJECT(data->treev_cards), "row-activated",
        G_CALLBACK(cb_treev_card_activated), data);


    /* Drag'n'drop */

    gtk_tree_view_enable_model_drag_source (data->treev_cards,
        GDK_BUTTON1_MASK, target_table, G_N_ELEMENTS(target_table),
        GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK);
    gtk_tree_view_enable_model_drag_dest (data->treev_categories,
        target_table, G_N_ELEMENTS(target_table),
        GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK);

    g_signal_connect (G_OBJECT(data->treev_cards), "drag_data_get",
        G_CALLBACK(cb_treev_cards_drag_data_get), data);
    g_signal_connect (G_OBJECT(data->treev_categories), "drag_data_received",
        G_CALLBACK(cb_treev_categories_drag_data_received), data);


    iconfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_PIXMAP, F_ICON, FALSE, NULL);

    if (iconfile != NULL) {
        gtk_window_set_icon_from_file (GTK_WINDOW(data->window),
            iconfile, NULL);
        g_free (iconfile);
    }

    refresh_category_list (data);
    app_window_refresh_card_view (data->treev_cards, data->ls_cards,
        file_get_current_category_cards (data->file));

    gtk_paned_set_position (GTK_PANED(data->hpaned),
        prefs_get_category_pane_width (data->prefs));

    gtk_tree_view_column_set_fixed_width (data->treev_col_card_front,
        prefs_get_card_column_front_width (data->prefs));

    app_window_select_treev_root (data->treev_categories);

    app_window_update_sensitivity (data);

    gtk_widget_show (data->window);
}

