/* ignuit - Educational software for the GNOME, following the Leitner
 * flash-card system.
 *
 * Copyright (C) 2008 Timothy Musson
 *
 * <trmusson@ihug.co.nz>
 * http://homepages.ihug.co.nz/~trmusson/programs.html#ignuit
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>
#include <gnome.h>
#include <glade/glade.h>

#include "main.h"
#include "prefs.h"
#include "file.h"
#include "dialog-editor.h"
#include "app-window.h"


typedef struct {

    File                *file;
    Prefs               *prefs;

    GladeXML            *gui;
    GtkWidget           *window;

    GtkTreeView         *treev;
    GtkListStore        *ls;

    GtkTextView         *textview;
    GtkTextBuffer       *textbuf;

    GtkWidget           *b_add;
    GtkWidget           *b_remove;
    GtkWidget           *b_prev;
    GtkWidget           *b_next;
    GtkWidget           *b_flip;

    GtkWidget           *m_add;
    GtkWidget           *m_remove;
    GtkWidget           *m_prev;
    GtkWidget           *m_next;
    GtkWidget           *m_flip;
    GtkWidget           *m_reset_stats;

    /* Make sure we save the correct side of the card if it's edited: */
    gboolean            ugly_flip_hack;

    GtkStatusbar        *statusbar;
    guint               sbcid;

} DialogData;


static DialogData *dialog = NULL;


static void
set_statusbar_message (DialogData *data, const gchar *text)
{
    gtk_statusbar_pop (data->statusbar, data->sbcid);
    gtk_statusbar_push (data->statusbar, data->sbcid, text);
}


static void
update_statusbar (DialogData *data)
{
	GList *item;
    gchar *s;

    item = file_get_current_item (data->file);
	if (!item) {
		s = g_strdup (_("No cards"));
	}
    else if (card_get_n_tests (CARD(item))) {
        float score;
		Card *c = CARD(item);
        score = card_get_n_known (c) / card_get_n_tests (c) * 100;
        s = g_strdup_printf (_("Group %d, Success %.0f%% (%d tests)"),
            card_get_group (c), score, card_get_n_tests (c));
    }
    else {
        s = g_strdup (_("Never tested"));
    }
    set_statusbar_message (data, s);
    g_free (s);
}


static void
set_interface_sensitivity (DialogData *data)
{
    GdkColor *color;
    GList    *item;
    Category *cat;
    gboolean have_fun;

    item = file_get_current_item (data->file);
    cat = file_get_current_category (data->file);
    have_fun = (cat && category_get_n_cards (cat) > 1);

    gtk_widget_set_sensitive (data->b_prev, have_fun);
    gtk_widget_set_sensitive (data->b_next, have_fun);
    gtk_widget_set_sensitive (data->m_prev, have_fun);
    gtk_widget_set_sensitive (data->m_next, have_fun);

    have_fun = (item != NULL);

    gtk_widget_set_sensitive (data->b_add, cat != NULL);
    gtk_widget_set_sensitive (data->b_remove, have_fun);
    gtk_widget_set_sensitive (data->b_flip, have_fun);
    gtk_widget_set_sensitive (data->m_add, cat != NULL);
    gtk_widget_set_sensitive (data->m_remove, have_fun);
    gtk_widget_set_sensitive (data->m_flip, have_fun);
    gtk_widget_set_sensitive (data->m_reset_stats, have_fun);

    gtk_text_view_set_editable (data->textview, have_fun);
    gtk_text_view_set_cursor_visible (data->textview, have_fun);

    color = prefs_get_color (data->prefs, COLOR_CARD_FG);
    gtk_widget_modify_text (GTK_WIDGET(data->textview),
        GTK_STATE_NORMAL, color);

    color = have_fun ? prefs_get_color (data->prefs, COLOR_CARD_BG) :
        prefs_get_color (data->prefs, COLOR_CARD_BG_END);
    gtk_widget_modify_base (GTK_WIDGET(data->textview),
        GTK_STATE_NORMAL, color);

    update_statusbar (data);
}


static void
display_text (GtkTextBuffer *textbuf, const gchar *text)
{
    if (text == NULL) {text = "";}

    gtk_text_buffer_set_text (textbuf, text, -1);
    gtk_text_buffer_set_modified(textbuf, FALSE);
}


static void
dialog_blank (void)
{
    display_text (dialog->textbuf, NULL);
    set_interface_sensitivity (dialog);
}


static void
show_current_card (DialogData *data)
{
    gboolean    flip;
    GList       *item;
    const gchar *text;


    /* Update textview */

    item = file_get_current_item (data->file);
    flip = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(data->b_flip));
    text = flip ? card_get_back (CARD(item)) : card_get_front (CARD(item));

    display_text (data->textbuf, text);


    /* Update button sensitivity and statusbar */

    set_interface_sensitivity (data);
}


static void
set_textview_properties (DialogData *data)
{
    GtkJustification just;
    const gchar *font;

    switch (file_get_style (data->file)) {
    case CARD_STYLE_HEADWORDS:
        font = prefs_get_font_large (data->prefs);
        just = GTK_JUSTIFY_CENTER;
        break;
    case CARD_STYLE_HEADWORDS_AND_EXAMPLES:
        font = prefs_get_font_small (data->prefs);
        just = GTK_JUSTIFY_CENTER;
        break;
    case CARD_STYLE_HEADWORDS_AND_SENTENCES:
    case CARD_STYLE_SENTENCES:
        font = prefs_get_font_small (data->prefs);
        just = GTK_JUSTIFY_LEFT;
        gtk_text_view_set_left_margin (data->textview, 6);
        gtk_text_view_set_right_margin (data->textview, 6);
        break;
    }

    if (font) {

        PangoFontDescription *desc;

        desc = pango_font_description_from_string (font);
        gtk_widget_modify_font (GTK_WIDGET(data->textview), desc);
        pango_font_description_free (desc);
    }

    gtk_text_view_set_justification (GTK_TEXT_VIEW(data->textview), just);
}


static gboolean
textview_check_changed (DialogData *data)
{
    /* Store any modified textview contents. */

    if (gtk_text_buffer_get_modified (data->textbuf)) {

        GtkTextIter start, end;
        gboolean    flip;
        GList       *item;
        gchar       *s;


        item = file_get_current_item (data->file);

        gtk_text_buffer_get_bounds (data->textbuf, &start, &end);
        s = gtk_text_buffer_get_text (data->textbuf, &start, &end, FALSE);

        flip = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(data->b_flip));
        if (data->ugly_flip_hack) { flip = !flip; }

        if (flip) {
            card_set_back (CARD(item), s);
        }
        else {
            card_set_front (CARD(item), s);
        }

        app_window_refresh_card_iter (data->treev, data->ls, item);

        file_set_changed (data->file, TRUE);
    }

    return file_get_changed (data->file);
}


static void
window_close (DialogData *data)
{
    gint w, h;

    textview_check_changed (data);

    gtk_window_get_size (GTK_WINDOW(data->window), &w, &h);
    prefs_set_quiz_size (data->prefs, w, h);

    gtk_widget_destroy (data->window);

    g_object_unref (G_OBJECT(data->gui));
    g_free (data);

    dialog = NULL;
}


static void
cb_m_add (GtkWidget *widget, DialogData *data)
{
    GtkTreeIter iter;
    GList       *item;
    Card        *c;


    /* Make a new blank card, add it to the currently selected category,
     * and set it up for editing.
     */

    textview_check_changed (data);

    c = card_new ();
    card_set_front (c, "");
    card_set_back (c, "");
    file_add_card (data->file, file_get_current_category (data->file), c);
    file_set_changed (data->file, TRUE);

    item = file_get_current_category_cards (data->file);
    file_set_current_item (data->file, item);

    app_window_add_card_iter (data->treev, data->ls, &iter, item);

    show_current_card (data);
}


static void
cb_m_remove (GtkWidget *widget, DialogData *data)
{
    GtkTreeIter iter;
    GList *search, *item, *next;


    if (!ask_yes_or_no (GTK_WINDOW(data->window),
        _("<b>Delete this card?</b>"),
        GTK_RESPONSE_NO)) {
        return;
    }

    /* Remove the current card from the main window's TreeView. */

    search = file_get_current_item (data->file);
    item = app_window_find_card_iter (data->treev, &iter, search);
    gtk_list_store_remove (data->ls, &iter);


    /* Remove the card from the file, and select the next card. */

    next = item->next ? item->next : item->prev;

    file_set_current_item (data->file, next);
    file_remove_card (data->file, CARD(item));
    file_set_changed (data->file, TRUE);


    /* Update or close the editor window. */

    if (next != NULL) {

        app_window_select_card (data->treev, next);

        if (!prefs_get_sticky_flips (data->prefs)) {
            gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(data->b_flip),
                FALSE);
        }
        show_current_card (data);
        return;
    }

    dialog_blank ();
}


static void
cb_m_reset_stats (GtkWidget *widget, DialogData *data)
{
    GList *item;


    /* Forget dates and counts for this card. */

    item = file_get_current_item (data->file);
    g_assert (item != NULL);

    card_reset_statistics (CARD(item));
    file_set_changed (data->file, TRUE);

    app_window_refresh_card_iter (data->treev, data->ls, item);

    update_statusbar (data);
}


static void
cb_m_prev (GtkWidget *widget, DialogData *data)
{
    textview_check_changed (data);

    if (!prefs_get_sticky_flips (data->prefs)) {
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(data->b_flip),
            FALSE);
    }

    file_decr_current_item (data->file, TRUE);
    app_window_select_card (data->treev, file_get_current_item (data->file));

    show_current_card (data);
}


static void
cb_m_next (GtkWidget *widget, DialogData *data)
{
    textview_check_changed (data);

    if (!prefs_get_sticky_flips (data->prefs)) {
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(data->b_flip),
            FALSE);
    }

    file_incr_current_item (data->file, TRUE);
    app_window_select_card (data->treev, file_get_current_item (data->file));

    show_current_card (data);
}


static void
cb_b_flip (GtkWidget *widget, DialogData *data)
{
    data->ugly_flip_hack = TRUE;
    textview_check_changed (data);
    data->ugly_flip_hack = FALSE;

    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM(data->m_flip),
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget))); 

    show_current_card (data);
}


static void
cb_m_flip (GtkWidget *widget, DialogData *data)
{
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(data->b_flip),
        gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM(widget))); 
}


static void
cb_m_sticky_flips (GtkWidget *widget, DialogData *data)
{
    prefs_set_sticky_flips (data->prefs,
        gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM(widget)));
}


static void
cb_m_close (GtkWidget *widget, DialogData *data)
{
    window_close (data);
}


static void
cb_window_delete (GtkWidget *widget, GdkEvent *event, DialogData *data)
{
    window_close (data);
}


void
dialog_editor_check_changed (void)
{
    if (dialog) {
        textview_check_changed (dialog);
    }
}


void
dialog_editor_kill (void)
{
    if (dialog) {
        window_close (dialog);
    }
}


void
dialog_editor_tweak (void)
{
    if (!dialog) { return; }

    if (!file_get_current_item (dialog->file)) {
        dialog_blank ();
        return;
    }

    show_current_card (dialog);
}


void
dialog_editor (Ignuit *ig, GtkWidget *parent,
    GtkTreeView *treev, GtkListStore *ls)
{
    GtkWidget  *m_close;
    GtkWidget  *m_sticky_flips;
    gchar      *gladefile;


    if (dialog != NULL) {
        gtk_window_present (GTK_WINDOW(dialog->window));
        show_current_card (dialog);
        return;
    }

    gladefile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_GLADE_EDITOR, TRUE, NULL);

    if (gladefile == NULL) {
        g_warning ("Can't find file: %s\n", F_GLADE_EDITOR);
        return;
    }

    dialog = g_new0 (DialogData, 1);
    dialog->gui = glade_xml_new (gladefile, NULL, NULL);
    g_free (gladefile);

    dialog->file = ig->file;
    dialog->prefs = ig->prefs;

    dialog->treev = treev;
    dialog->ls = ls;

    dialog->window = glade_xml_get_widget (dialog->gui, "dialog");

    gtk_window_set_default_size (GTK_WINDOW(dialog->window),
        prefs_get_quiz_width (dialog->prefs),
        prefs_get_quiz_height (dialog->prefs));

    dialog->textview = GTK_TEXT_VIEW(glade_xml_get_widget
        (dialog->gui, "textview"));
    dialog->textbuf = gtk_text_view_get_buffer (dialog->textview);
    dialog->ugly_flip_hack = FALSE;

    dialog->m_add = glade_xml_get_widget (dialog->gui, "m_add");
    dialog->m_remove = glade_xml_get_widget (dialog->gui, "m_remove");
    dialog->m_reset_stats = glade_xml_get_widget (dialog->gui, "m_reset_stats");
    dialog->m_prev = glade_xml_get_widget (dialog->gui, "m_prev");
    dialog->m_next = glade_xml_get_widget (dialog->gui, "m_next");
    dialog->m_flip = glade_xml_get_widget (dialog->gui, "m_flip");
    m_sticky_flips = glade_xml_get_widget (dialog->gui, "m_sticky_flips");
    m_close = glade_xml_get_widget (dialog->gui, "m_close");

    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM(m_sticky_flips),
        prefs_get_sticky_flips (dialog->prefs));

    dialog->b_add = glade_xml_get_widget (dialog->gui, "b_add");
    dialog->b_remove = glade_xml_get_widget (dialog->gui, "b_remove");
    dialog->b_prev = glade_xml_get_widget (dialog->gui, "b_prev");
    dialog->b_next = glade_xml_get_widget (dialog->gui, "b_next");
    dialog->b_flip = glade_xml_get_widget (dialog->gui, "b_flip");

    dialog->statusbar = (GtkStatusbar*)glade_xml_get_widget
        (dialog->gui, "statusbar");
    dialog->sbcid = gtk_statusbar_get_context_id (dialog->statusbar, "editor");

    g_signal_connect (G_OBJECT(dialog->window), "delete-event",
        G_CALLBACK(cb_window_delete), dialog);

    g_signal_connect (G_OBJECT(dialog->m_add), "activate",
        G_CALLBACK(cb_m_add), dialog);
    g_signal_connect (G_OBJECT(dialog->m_remove), "activate",
        G_CALLBACK(cb_m_remove), dialog);
    g_signal_connect (G_OBJECT(dialog->m_reset_stats), "activate",
        G_CALLBACK(cb_m_reset_stats), dialog);
    g_signal_connect (G_OBJECT(dialog->m_prev), "activate",
        G_CALLBACK(cb_m_prev), dialog);
    g_signal_connect (G_OBJECT(dialog->m_next), "activate",
        G_CALLBACK(cb_m_next), dialog);
    g_signal_connect (G_OBJECT(dialog->m_flip), "toggled",
        G_CALLBACK(cb_m_flip), dialog);
    g_signal_connect (G_OBJECT(m_sticky_flips), "toggled",
        G_CALLBACK(cb_m_sticky_flips), dialog);
    g_signal_connect (G_OBJECT(m_close), "activate",
        G_CALLBACK(cb_m_close), dialog);

    g_signal_connect (G_OBJECT(dialog->b_add), "clicked",
        G_CALLBACK(cb_m_add), dialog);
    g_signal_connect (G_OBJECT(dialog->b_remove), "clicked",
        G_CALLBACK(cb_m_remove), dialog);
    g_signal_connect (G_OBJECT(dialog->b_prev), "clicked",
        G_CALLBACK(cb_m_prev), dialog);
    g_signal_connect (G_OBJECT(dialog->b_next), "clicked",
        G_CALLBACK(cb_m_next), dialog);
    g_signal_connect (G_OBJECT(dialog->b_flip), "toggled",
        G_CALLBACK(cb_b_flip), dialog);

    gtk_window_set_transient_for (GTK_WINDOW(dialog->window),
        GTK_WINDOW(parent));

    gtk_window_set_modal (GTK_WINDOW(dialog->window), FALSE);


    set_textview_properties (dialog);
    show_current_card (dialog);

    gtk_widget_show (dialog->window);
}

